import numpy as np
import tensorflow as tf


@tf.keras.utils.register_keras_serializable('PositionalEncoding')
class PositionalEncoding(tf.keras.layers.Layer):
    """
    Implements functionality for additive periodic positional encoding.
    Directly adapted from https://www.tensorflow.org/text/tutorials/transformer.
    """

    def __init__(self, seqlen, d_model, num_dims=3, **kwargs):
        """
        Initialize the positional encoding layer.
        Assumes the input has shape (batch, blocks, inner) or (batch, _ , blocks, inner)
        :param seqlen: length of the sequence = number of blocks
        :param d_model: dimension of the encoding (needs to match block size)
        :param num_dims: number of input dimensions to enable easy broadcasting
        """
        super().__init__(**kwargs)
        self.seqlen = seqlen
        self.d_model = d_model
        self.num_dims = num_dims
        self.pos_encoding = self._positional_encoding(length=seqlen, depth=d_model)

    def _positional_encoding(self, length, depth):
        """
        Generate the constant matrices representing the encoding
        :param length: number of blocks
        :param depth: block dimension
        :return:
        """
        depth = depth / 2

        positions = np.arange(length)[:, np.newaxis]  # (seq, 1)
        depths = np.arange(depth)[np.newaxis, :] / depth  # (1, depth)

        angle_rates = 1 / (10000**depths)  # (1, depth)
        angle_rads = positions * angle_rates  # (pos, depth)

        pos_encoding = np.concatenate([np.sin(angle_rads), np.cos(angle_rads)], axis=-1)

        return tf.constant(tf.cast(pos_encoding, dtype=tf.float32))

    def call(self, x):
        """
        Performs a (symbolic) forward pass
        :param x: tf.tensor of shape (..., n_blocks, block_size)
        :return: tf.tensor of same shape
        """
        # This factor sets the relative scale of the embedding and positional_encoding.
        x *= tf.math.sqrt(tf.cast(self.d_model, tf.float32))
        if self.num_dims == 3:
            x = x + self.pos_encoding[tf.newaxis, tf.newaxis, :, :]
        else:
            x = x + self.pos_encoding[tf.newaxis, :, :]
        return x

    def get_config(self):
        """
        Helper function necessary to save model
        :return: dict config
        """
        config = super().get_config()
        config.update(
            {"seqlen": self.seqlen, "d_model": self.d_model, "num_dims": self.num_dims}
        )
        return config



""""""
import multiprocessing
from argparse import Namespace

import numpy as np
import tensorflow as tf

from beliefppg.datasets.file_reader import (
    load_bami_1,
    load_bami_2,
    load_dalia,
    load_ieee,
    load_wesad,
)
from beliefppg.model.config import InputConfig
from beliefppg.util.preprocessing import (
    get_strided_windows,
    process_window_spec_acc,
    process_window_spec_ppg,
    process_window_time,
)


def prepare_session_spec(
    ppg, acc, ppg_freq, acc_freq, win_size, stride, n_bins, min_hz, max_hz
):
    """
    Prepares the spectral-domain input. Performs a custom version of STFT with channel averaging.
    Only returns frequencies within a relevant range.
    :param ppg: the ppg signal with shape (n_samples, n_ppg_channels)
    :param acc: the acceleromenter signal with shape (n_samples, n_acc_channels)
    :param ppg_freq: frequency of the PPG signal
    :param acc_freq: frequency of the ACC signal
    :param win_size: size in seconds of the signal window for fourier transform
    :param stride: size in seconds of the window stride for STFT
    :param n_bins: number of frequency bins. Can be either 64 or 256 for model compatibility
    :param min_hz: minimal relevant frequency in Hz.
    :param max_hz: maximal relevant frequency in Hz.
    :return: processed signal of shape (n_steps, n_bins, n_channels),
            where n_channels=2 stands for the aggregated PPG and ACC signals
    """
    fft_winsize = (
        535 if n_bins == 64 else (4 * 535 - 5)
    )
    ppgs = []
    ppg_wsize = win_size * ppg_freq
    for i in range(0, len(ppg) - ppg_wsize + 1, ppg_freq * stride):
        ppgs.append(
            process_window_spec_ppg(
                ppg[i : i + ppg_wsize], ppg_freq, fft_winsize, min_hz, max_hz
            )
        )

    accs = []
    acc_wsize = win_size * acc_freq
    for i in range(0, len(acc) - acc_wsize + 1, acc_freq * stride):
        accs.append(
            process_window_spec_acc(
                acc[i : i + acc_wsize], acc_freq, fft_winsize, min_hz, max_hz
            )
        )

    sig = np.stack([ppgs, accs], axis=-1)

    # normalize
    sig = (sig - sig.mean()) / (sig.std() + 1e-10)

    assert not np.isnan(sig).any()
    return sig.astype(np.float32)


def prepare_session_time(ppg, ppg_freq, target_freq, filter_lowcut, filter_highcut):
    """
    Prepares the time-domain input. Preprocessing mainly consists of filtering, resampling and standardizing.
    :param ppg: the ppg signal with shape (n_samples, n_channels)
    :param ppg_freq: frequency of the signal
    :param target_freq: desired input frequency for model
    :param filter_lowcut: lower cutoff frequency for bandpass filter
    :param filter_highcut: upper cutoff frequency for bandpass filter
    :return: processed signal of shape (n_samples_new,)
    """
    # only feed ppg signals as time-domain features
    sig = process_window_time(ppg, ppg_freq, target_freq, filter_lowcut, filter_highcut)

    # normalize
    sig = (sig - sig.mean()) / (sig.std() + 1e-10)

    assert not np.isnan(sig).any()
    return sig.astype(np.float32)


def prepare_session_labels(hr, n_frames):
    """
    Add offset to session labels to compensate for the length of the feature window.
    :param hr: labels of shape (n_steps, )
    :param n_frames: number of steps per feature window
    :return: labels of shape (n_steps - n_frames + 1, )
    """
    offset = n_frames - 1
    assert not np.isnan(hr).any()
    return hr[offset:].astype(np.float32)


def get_sessions(args: Namespace):
    """
    Loads and preprocesses the sessions of a dataset into tf.data.Datasets containing
    input (spectral- and time-domain features) aligned with labels.
    :param args: command line arguments containing config
    :return: tuple of (sessions, session_names) where
            sessions is a list of tf.data.Datasets
                each yielding tuples of shape (((n_frames, n_bins, 2), (n_steps, 1)), 1)
                representing (input, labels) where input=(spec_feat, time_feat) and the label is the BPM HR value.
            session_names is a list of corresponding session descriptors
    """

    if args.dataset == "all":
        datasets = ["dalia", "wesad", "bami-1", "bami-2", "ieee"]
    elif args.dataset == "bami":
        datasets = ["bami-1", "bami-2"]
    else:
        datasets = [args.dataset]

    sessions = []
    names = []
    ppg_freqs = []
    acc_freqs = []

    for dataset in datasets:
        if dataset == "dalia":
            sessions_, names_, ppg_freq, acc_freq = load_dalia(args.data_dir)
        elif dataset == "wesad":
            sessions_, names_, ppg_freq, acc_freq = load_wesad(args.data_dir)
        elif dataset == "bami-1":
            sessions_, names_, ppg_freq, acc_freq = load_bami_1(args.data_dir)
        elif dataset == "bami-2":
            sessions_, names_, ppg_freq, acc_freq = load_bami_2(args.data_dir)
        elif dataset == "ieee":
            sessions_, names_, ppg_freq, acc_freq = load_ieee(args.data_dir)
        else:
            raise NotImplementedError(r"Dataset %s not supported (yet)" % args.dataset)

        sessions.extend(sessions_)
        names.extend(names_)
        ppg_freqs.extend([ppg_freq] * len(sessions_))
        acc_freqs.extend([acc_freq] * len(sessions_))

    window_len = InputConfig.WINSIZE  # hard-code window length in seconds to match the labels
    stride = InputConfig.STRIDE  # window stride in seconds

    pool = multiprocessing.Pool()

    # compute time-frequency features
    spectral_feat = pool.starmap(
        prepare_session_spec,
        (
            (
                ppg,
                acc,
                ppg_freq,
                acc_freq,
                window_len,
                stride,
                args.n_bins,
                args.min_hz,
                args.max_hz,
            )
            for ((ppg, acc, hr), ppg_freq, acc_freq) in zip(sessions,
                                                            ppg_freqs,
                                                            acc_freqs)
        ),
    )
    spectral_dss = map(tf.data.Dataset.from_tensor_slices, spectral_feat)
    spectral_dss = [
        get_strided_windows(sesh, win_size=args.n_frames, stride=1)
        for sesh in spectral_dss
    ]

    # compute time-domain features
    time_feat = pool.starmap(
        prepare_session_time,
        (
            (ppg, ppg_freq, args.freq, args.filter_lowcut, args.filter_highcut)
            for ((ppg, acc, hr), ppg_freq) in zip(sessions,
                                                  ppg_freqs)
        ),
    )
    time_winsize = window_len + (args.n_frames - 1) * stride
    time_dss = map(tf.data.Dataset.from_tensor_slices, time_feat)
    time_dss = [
        get_strided_windows(
            ds, win_size=time_winsize * args.freq, stride=stride * args.freq
        )
        for ds in time_dss
    ]

    # align labels
    labels = [prepare_session_labels(hr, args.n_frames) for (ppg, acc, hr) in sessions]
    label_dss = map(tf.data.Dataset.from_tensor_slices, labels)

    # create datasets yielding tuples of form ((X_spec, X_time), y)
    X_dss = map(tf.data.Dataset.zip, zip(spectral_dss, time_dss))
    joint_dss = map(tf.data.Dataset.zip, zip(X_dss, label_dss))

    sessions = np.asarray([ds.prefetch(tf.data.AUTOTUNE) for ds in joint_dss])
    return sessions[np.argsort(names)], sorted(names)


def join_sessions(dss, shuffle):
    """
    Combines a list of per-session tf datasets. Interleaves them (otherwise we would need a large shuffle buffer)
    :param dss: list of tf.data.Datasets
    :param shuffle: whether to shuffle
    :return: concatenated or randomly interleaved tf dataset
    """
    if shuffle:
        dss = [ds.shuffle(1000) for ds in dss]
        # interleave datasets
        comb = tf.data.Dataset.from_tensor_slices(dss).interleave(
            lambda x: x,
            cycle_length=len(dss),
            num_parallel_calls=tf.data.AUTOTUNE,
            deterministic=False,
        )
    else:
        comb = dss[0]
        for i in range(1, len(dss)):
            comb = comb.concatenate(dss[i])

    return comb




import numpy as np
from tensorflow.keras.models import (
    Model,
    load_model,
    model_from_json
    )
from tensorflow.keras.utils import custom_object_scope

import beliefppg.model.belief_ppg
from beliefppg.model.belief_ppg import (
    AveragePooling1D,
    FlexibleAttention,
    )
from beliefppg.model.binned_regression_loss import BinnedRegressionLoss
from beliefppg.model.positional_encoding import PositionalEncoding
from beliefppg.model.prior_layer import PriorLayer


def load_inference_model(model_path: str) -> Model:
    """
    Load a model from a given path
    :param model_path: path to the model
    :return: keras.models.Model
    """
    custom_objects = {'AveragePooling1D': AveragePooling1D,
                      'BinnedRegressionLoss': BinnedRegressionLoss,
                      'PositionalEncoding': PositionalEncoding,
                      'FlexibleAttention': FlexibleAttention,
                      'PriorLayer': PriorLayer}

    with custom_object_scope(custom_objects):
        with open(model_path + '_architecture.json', 'r') as json_file:
            loaded_model_json = json_file.read()
        model = model_from_json(loaded_model_json)

        model_weights = np.load(model_path + '_weights.npz')
        weights_list = [model_weights[key] for key in model_weights]
        model.set_weights(weights_list)

        return model



from argparse import Namespace

import tensorflow as tf
from tensorflow.keras.layers import Layer
from tensorflow.keras.layers import (
    Add,
    Activation,
    Attention,
    Concatenate,
    Conv1D,
    Conv2D,
    Dense,
    Dropout,
    Flatten,
    MaxPooling1D,
    Multiply,
    Permute,
    Reshape,
    UpSampling1D,
    )

from beliefppg.model.config import InputConfig
from beliefppg.model.positional_encoding import PositionalEncoding
from beliefppg.model.timedomain_backbone import get_timedomain_backbone


@tf.keras.utils.register_keras_serializable('AveragePooling1D')
class AveragePooling1D(Layer):
    def __init__(self, dim, **kwargs):
        super().__init__(**kwargs)
        self.dim = dim

    def call(self, inputs):
        return tf.reduce_mean(inputs, axis=self.dim)

    def get_config(self):
        config = super().get_config()
        config["dim"] = self.dim
        return config


@tf.keras.utils.register_keras_serializable('FlexibleAttention')
class FlexibleAttention(Layer):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.attention = Attention()

    def build(self, input_shape):
        # Check if the input is a list of tensors
        if isinstance(input_shape, list):
            # Assume each item in the list has the same shape
            self.use_reshaping = len(input_shape[0]) > 3
            if self.use_reshaping:
                # Compute intermediate and original shapes for the first tensor in the list
                self.intermediate_shape = (-1, input_shape[0][-2], input_shape[0][-1])
                self.original_shape = [-1,] + [i for i in input_shape[0][1:-2]] + [input_shape[0][-2], input_shape[0][-1]]
        else:
            self.use_reshaping = len(input_shape) > 3
            if self.use_reshaping:
                self.intermediate_shape = (-1, input_shape[-2], input_shape[-1])
                self.original_shape = [-1,] + [i for i in input_shape[1:-2]] + [input_shape[-2], input_shape[-1]]
        super().build(input_shape)

    def call(self, inputs):
        if not isinstance(inputs, list):
            inputs = [inputs, inputs]  # For self-attention if a single tensor is provided

        if self.use_reshaping:
            reshaped_inputs = [tf.reshape(input_tensor, self.intermediate_shape) for input_tensor in inputs]
            attention_output = self.attention(reshaped_inputs)
            output = tf.reshape(attention_output, self.original_shape)
        else:
            output = self.attention(inputs)

        return output

    def compute_output_shape(self, input_shape):
        if isinstance(input_shape, list):
            return input_shape[0]
        return input_shape


def attention_block_1d(x, g, inter_channel):
    """
    helper function for attention block used in U-Net (upsampling path)
    Based on https://github.com/lixiaolei1982/Keras-Implementation-of-U-Net-R2U-Net-Attention-U-Net-Attention-R2U-Net.-
    :param x: branch 1 (from bottleneck)
    :param g: branch 2 (from down)
    :param inter_channel: nb channels to use internally
    :return: joined branches, same shape
    """
    # theta_x(?,g_height,g_width,inter_channel)

    theta_x = Conv1D(inter_channel, 1, strides=1, data_format="channels_last")(x)

    # phi_g(?,g_height,g_width,inter_channel)

    phi_g = Conv1D(inter_channel, 1, strides=1, data_format="channels_last")(g)

    # f(?,g_height,g_width,inter_channel)

    f = Activation("relu")(Add()([theta_x, phi_g]))

    # psi_f(?,g_height,g_width,1)

    psi_f = Conv1D(1, 1, strides=1, data_format="channels_last")(f)

    rate = Activation("sigmoid")(psi_f)

    # rate(?,x_height,x_width)

    # att_x(?,x_height,x_width,x_channel)
    att_x = Multiply()([x, rate])

    return att_x


def attention_up_and_concate(
    down_layer, layer, data_format="channels_first", down_fac=2
):
    """
    helper function to define an upsampling block
    Based on https://github.com/lixiaolei1982/Keras-Implementation-of-U-Net-R2U-Net-Attention-U-Net-Attention-R2U-Net.-
    :param down_layer: skip connections
    :param layer: upsampling branch
    :param data_format: where the channels are
    :param down_fac: how much to down/upsample each depth level
    :return: output of depth level
    """
    if data_format == "channels_first":
        in_channel = down_layer.shape[1]
    else:
        in_channel = down_layer.shape[-1]

    up = UpSampling1D(size=down_fac)(down_layer)

    layer = attention_block_1d(x=layer, g=up, inter_channel=in_channel // 4)

    if data_format == "channels_first":
        combined = Concatenate(axis=1)([up, layer])
    else:
        combined = Concatenate(axis=-1)([up, layer])

    return combined


def double_attn(inp, n_frames, n_bins, channels):
    """
    Constructs a self-attention block that applies one-dimensional attention across frames and bins in parallel.
    :param inp: tf.tensor of shape (batch, n_frames, n_bins, _)
    :param n_frames: first dimension
    :param n_bins: second dimension
    :param channels: last dimension
    :return: tuple containing two tf.tensor's (time_result, channel_result)
                each of shape (batch, n_frames, n_bins, channels)
    """
    x1 = Conv2D(channels, (3, 3), padding="same", activation="leaky_relu")(inp)
    x1 = Dropout(0.1)(x1)
    x1 = Conv2D(channels, (3, 3), padding="same", activation="leaky_relu")(x1)
    x1 = Dropout(0.1)(x1)

    # self-attention over time
    x1_freq = PositionalEncoding(seqlen=n_bins, d_model=channels)(x1)
    freq_attn = FlexibleAttention()([Dense(channels)(x1_freq), Dense(channels)(x1_freq)])

    # self-attention over frequency space
    x1_time = PositionalEncoding(seqlen=n_frames, d_model=channels)(
        Permute((2, 1, 3))(x1)
    )
    time_attn = FlexibleAttention()([Dense(channels)(x1_time), Dense(channels)(x1_time)])
    time_attn = Permute((2, 1, 3))(time_attn)

    return time_attn, freq_attn


def hybrid_unet(
    spec_input,
    time_input,
    depth=3,
    attn_channels=32,
    init_channels=12,
    down_fac=4,
    use_time_backbone=True,
):
    """
    Builds the architecture at the core of the model, involving
    1) an attention block followed by the reduction of the time dimension
    2) an U-Net performing frequency-to-frequency estimation
    3) a time-domain backbone consisting of a CNN-LSTM that is attached to the U-Net bottleneck.
    Note that the number of frequency bins needs to be compatible with the downsampling factor, i.e. divisible by 4
    :param spec_input: tf.tensor of shape (batch_size, n_frames, n_bins, 2)
    :param time_input: tf.tensor of shape (batch_size, n_timesteps , 1)
    :param depth: number of contraction steps in the U-Net
    :param attn_channels: inner dimension for initial attention block
    :param init_channels: channels to start from in first layer of U-Net. Are doubled in each contraction step.
    :param down_fac: downsampling factor of the U-Net
    :param use_time_backbone: whether to attach the time-domain backbone to the NN or not
    :return: tf.tensor of shape (batch_size, n_bins)
    """
    time_attn, freq_attn = double_attn(
        spec_input, spec_input.shape[1], spec_input.shape[2], attn_channels
    )
    # reduce mean over 7 time steps
    x = Dense(attn_channels)(spec_input) + time_attn + freq_attn
    x = AveragePooling1D(1)(x)
    # resulting dimension (batch, freq_bins, channels)

    # U-net code until bottleneck
    skips = []
    channels = init_channels  # channels to start with

    # down
    for i in range(depth):
        x = Conv1D(
            channels, 3, activation="relu", padding="same", data_format="channels_last"
        )(x)
        x = Dropout(0.2)(x)
        x = Conv1D(
            channels, 3, activation="relu", padding="same", data_format="channels_last"
        )(x)
        skips.append(x)
        x = MaxPooling1D(down_fac, strides=down_fac, data_format="channels_last")(x)
        channels = channels * 2

    if use_time_backbone:
        # get backbone branches
        weight_branch, feat_branch = get_timedomain_backbone(time_input, x.shape[-1])

        # concatenate them with bottleneck
        shape = weight_branch.shape
        reshape_layer = Reshape([1, shape[-1]])

        weight_branch = Concatenate(axis=-2)([reshape_layer(weight_branch), x])
        feat_branch = Concatenate(axis=-2)([reshape_layer(feat_branch), x])

        # convolve weight back to shape (1,n_channels), then normalize
        weight_branch = Conv1D(x.shape[-1], 2, activation="tanh")(weight_branch)
        weight_branch = Dropout(0.2)(weight_branch)

        # also convolve feat
        feat_branch = Conv1D(x.shape[-1], 2, activation="relu")(feat_branch)
        feat_branch = Dropout(0.2)(feat_branch)

        # combine
        x = x + weight_branch * feat_branch

    # up
    for i in reversed(range(depth)):
        channels = channels // 2
        x = attention_up_and_concate(
            x, skips[i], data_format="channels_last", down_fac=down_fac
        )
        x = Conv1D(
            channels, 3, activation="relu", padding="same", data_format="channels_last"
        )(x)
        x = Dropout(0.2)(x)
        x = Conv1D(
            channels, 3, activation="relu", padding="same", data_format="channels_last"
        )(x)

    conv6 = Conv1D(1, 1, padding="same", data_format="channels_last")(x)

    # out
    x = Flatten()(conv6)
    return x


def build_belief_ppg(n_frames: int, n_bins: int, freq: int, use_time_backbone=True) -> tf.keras.models.Model:
    """
    Wrapper function constructing the BeliefPPG architecture

    :param n_frames: The number of frames to consider in the spectrogram input.
    :param n_bins: The number of frequency bins in each frame of the spectrogram.
    :param freq: The frequency resolution of the time input.
    :return: tf.kers.models.Model the uncompiled functional model
    """
    spec_input = tf.keras.layers.Input(shape=(n_frames, n_bins, 2))
    time_input = tf.keras.layers.Input(
        shape=(freq * (n_frames - 1) * InputConfig.STRIDE + freq * InputConfig.WINSIZE, 1)
    )

    logits = hybrid_unet(spec_input, time_input, use_time_backbone=use_time_backbone)

    out = Activation("softmax")(logits)
    model = tf.keras.models.Model(inputs=[spec_input, time_input], outputs=out)
    return model



import os
from typing import (
    Optional,
    Tuple,
)

import numpy as np
import tensorflow as tf

from beliefppg.datasets.pipeline_generator import (
    prepare_session_spec,
    prepare_session_time,
)
from beliefppg.model.config import InputConfig
from beliefppg.model.load import load_inference_model
from beliefppg.util.preprocessing import get_strided_windows


def infer_hr_uncertainty(ppg: np.array, ppg_freq: int, acc: Optional[np.ndarray] = None, acc_freq: Optional[int] = None,
             decoding: str = "sumproduct", use_time_backbone=True, uncertainty: str="entropy",
             batch_size: int = 128, filter_lowcut: float = 0.1, filter_highcut: float = 18.0,
             use_gpu: bool = False, model_path: str = None) -> Tuple[np.array, np.array, np.array]:
    """
    Infers heart rate from PPG and accelerometer data using the specified decoding method and returning uncertainty.
    :param ppg: PPG signal data with shape (n_samples, n_channels).
    :param ppg_freq: Sampling frequency of the PPG signal in Hz
    :param acc: Accelerometer signal data with shape (n_samples, n_channels). BeliefPPG to function without accelerometer signal data, but its accuracy may be reduced.
    :param acc_freq: Sampling frequency of the accelerometer signal in Hz
    :param decoding: Decoding method to use, either "sumproduct" or "viterbi"
    :param use_time_backbone: Whether to use the time-domain backbone or not
    :param uncertainty: Metric for predictive uncertainty, either "entropy" or "std"
    :param batch_size: Batch size for inference
    :param filter_lowcut: Lowcut frequency for filtering (per default set to 0.1 Hz which is used for training the default model)
    :param filter_highcut: Highcut frequency for filtering (per default set to 18.0 Hz which is used for training the default model)
    :param use_gpu: Whether to use GPU for inference or not
    :param model_path: Path to the inference model. If None, the default model will be loaded.
    :return: Tuple of predicted heart rates [BPM], uncertainties, and time intervals [s]
    """

    if ppg.ndim != 2:
        raise ValueError("PPG signal data must have shape (n_samples, n_channels)")

    if acc is None:
        acc = np.zeros((ppg.shape[0], 3))
        acc_freq = ppg_freq
        print("Warning: No accelerometer data provided. Estimation accuracy may be reduced.")
    elif acc_freq is None:
        raise ValueError("Accelerometer frequency must be provided if accelerometer data is provided.")
    elif acc.ndim != 2:
        raise ValueError("Accelerometer signal data must have shape (n_samples, n_channels)")

    if not isinstance(ppg_freq, int):
        print("Warning: ppg_freq is not an integer, converting to integer.")
        ppg_freq = round(ppg_freq)
    if not isinstance(acc_freq, int):
        print("Warning: acc_freq is not an integer, converting to integer.")
        acc_freq = round(acc_freq)

    # Set TensorFlow to use GPU or CPU based on the parameter
    if use_gpu:
        physical_devices = tf.config.list_physical_devices('GPU')
        if physical_devices:
            tf.config.experimental.set_memory_growth(physical_devices[0], True)
        else:
            print("No GPU devices found, switching to CPU.")
            os.environ['CUDA_VISIBLE_DEVICES'] = '-1'
    else:
        os.environ['CUDA_VISIBLE_DEVICES'] = '-1'

    # Load default model if model_path is not provided
    if model_path is None:
        model_dir = os.path.dirname(os.path.abspath(__file__))
        if use_time_backbone:
            model_path = os.path.join(model_dir, "inference_model")
        else:
            model_path = os.path.join(model_dir, "inference_model_notimebackbone")

    # Load the inference model
    inference_model = load_inference_model(model_path)
    prior_layer = inference_model.layers[-1]
    min_hz = prior_layer.min_hz
    max_hz = prior_layer.max_hz
    n_bins = prior_layer.dim
    n_frames = inference_model.input_spec[0].shape[1]
    win_size = InputConfig.WINSIZE
    stride = InputConfig.STRIDE
    time_winsize = win_size + (n_frames - 1) * stride
    freq = inference_model.input_spec[1].shape[1] // time_winsize

    # Set up the decoding method
    if decoding == "sumproduct":
        prior_layer = inference_model.get_layer("prior_layer")
        prior_layer.set_online(True)
    elif decoding == "viterbi":
        prior_layer = inference_model.get_layer("prior_layer")
        prior_layer.set_online(False)
    else:
        raise NotImplementedError(f"Decoding method {decoding} not implemented")

    # Configure uncertainty output
    prior_layer = inference_model.get_layer("prior_layer")
    prior_layer.set_uncertainty(uncertainty)

    # Prepare spectral features
    spectral_feat = prepare_session_spec(ppg, acc, ppg_freq, acc_freq,
                                         win_size, stride, n_bins, min_hz,
                                         max_hz)
    spectral_dss = tf.data.Dataset.from_tensor_slices(spectral_feat)
    spectral_dss = get_strided_windows(spectral_dss, win_size=n_frames, stride=1)

    # Prepare time-domain features
    time_feat = prepare_session_time(ppg, ppg_freq, freq,
                                     filter_lowcut, filter_highcut)
    time_dss = tf.data.Dataset.from_tensor_slices(time_feat)
    time_dss = get_strided_windows(time_dss, win_size=time_winsize*freq,
                                   stride=stride*freq)

    # Combine spectral and time-domain features into a single dataset
    X_dss = tf.data.Dataset.zip(spectral_dss, time_dss)

    # Create dummy labels for the dataset
    labels = [-1 for _ in time_dss]
    label_dss = tf.data.Dataset.from_tensor_slices(labels)
    joint_dss = tf.data.Dataset.zip(X_dss, label_dss)

    # Batch and prefetch the dataset for inference
    batches = joint_dss.batch(batch_size).prefetch(tf.data.AUTOTUNE)
    y_pred, uncertainty = inference_model.predict(batches)

    # Compute time intervals
    time_start = np.arange(0, len(y_pred) * stride, stride, dtype=np.float32)
    time_intervals = np.column_stack((time_start, time_start + stride))

    return y_pred, uncertainty, time_intervals


def infer_hr(ppg: np.array, ppg_freq: int, acc: Optional[np.ndarray] = None, acc_freq: Optional[int] = None) -> Tuple[np.array, np.array]:
    """
    Infers heart rate from PPG and accelerometer data.
    :param ppg: PPG signal data with shape (n_samples, n_channels).
    :param ppg_freq: Sampling frequency of the PPG signal in Hz
    :param acc: Accelerometer signal data with shape (n_samples, n_channels). BeliefPPG to function without accelerometer signal data, but its accuracy may be reduced.
    :param acc_freq: Sampling frequency of the accelerometer signal in Hz
    :return: Tuple of predicted heart rates [BPM], and time indices [samples] at midpoints of the windows used for HR inference 
    """

    y_pred, uncertainty, time_intervals = infer_hr_uncertainty(ppg, ppg_freq, acc, acc_freq)
    midpoint_idxs = (np.mean(time_intervals, axis=-1)*ppg_freq).astype(int)
    return y_pred, midpoint_idxs



import tensorflow as tf
import tensorflow_probability as tfp


class BinnedRegressionLoss(tf.keras.losses.Loss):
    """
    Implements the discretized regression loss function.
    """

    def __init__(
        self,
        dim: int,
        min_hz: float,
        max_hz: float,
        sigma_y: float,
        name="binned_log_likelihood",
        **kwargs
    ):
        """
        Initializes the layer from config
        :param dim: number of output bins of the model
        :param min_hz: lowest predictable HR value in Hz
        :param max_hz: highest predictable HR value in Hz
        :param sigma_y: standard deviation of gaussian we model ground truth HR [distribution] with
        :param name: name of the node
        :param kwargs:
        """
        super().__init__(name=name, **kwargs)
        self.sigma_y = sigma_y
        self.min_hz = min_hz
        self.max_hz = max_hz
        min_bpm = min_hz * 60
        max_bpm = max_hz * 60
        self.dim = dim
        self.bin_edges = tf.range(
            min_bpm, max_bpm, (max_bpm - min_bpm) / self.dim
        )

    @tf.function
    def y_to_bins(self, y):
        """
        creates binned representation of continuous label
        -> Gaussian-Bump representation around true HR
        :param y: label in BPM of shape (n_steps, )
        :return: target bin probabilities of shape (n_steps, self.dim)
        """
        distr = tfp.distributions.Normal(y, self.sigma_y)
        prob = distr.prob(self.bin_edges)
        return prob / tf.reduce_sum(prob)

    def call(self, y_true, y_pred):
        """
        performs a (symbolic) forward pass:
         1) Converts y to binned probability density representation (a normal distr. around the target)
         2) Calculates the cross-entropy between predicted and target distribution
        :param y_true: label in BPM of shape (n_steps,)
        :param y_pred: predictions for each bin o shape (n_steps, self.dim)
        :return:
        """
        if len(y_true.shape) == 1:
            y_true = tf.expand_dims(y_true, -1)

        binned_true = self.y_to_bins(y_true)
        losses = tf.keras.metrics.categorical_crossentropy(binned_true, y_pred, axis=-1)
        return tf.reduce_sum(losses)

    def get_config(self):
        """
        Helper function necessary to save model
        :return: dict config
        """
        config = {
            "sigma_y": self.sigma_y,
            "min_hz": self.min_hz,
            "max_hz": self.max_hz,
            "dim": self.dim,
        }
        base_config = super().get_config()
        return {**base_config, **config}



import numpy as np
import scipy
from scipy.fft import rfft, rfftfreq
from scipy.signal import butter, lfilter


def get_strided_windows(ds, win_size, stride):
    """
    Generates a sliding window view of a tf.data.Dataset along the first axis
    :param win_size: number of samples in window
    :param stride: number of skipped samples between consecutive windows
    :return: tf.data.Dataset yielding tensors of shape (win_size, ) + old_shape
    """
    ds = ds.window(size=win_size, shift=stride, drop_remainder=True)
    return ds.flat_map(lambda window: window.batch(win_size))


def butter_bandpass(lowcut, highcut, fs, order=4):
    """
    Helper function for butter_bandpass_filter
    Creates a scipy filter
    https://stackoverflow.com/questions/12093594/how-to-implement-band-pass-butterworth-filter-with-scipy-signal-butter

    :param lowcut: lower cut-off value for butterworth filter (Hz)
    :param highcut: upper cut-off value for butterworth filter (Hz)
    :param fs: frequency of signal in Hz
    :param order: order of filter
    :return:(ndarray, ndarray)  Numerator (`b`) and denominator (`a`) polynomials of the IIR filter.scipy filter object
    """
    nyq = 0.5 * fs
    low = lowcut / nyq
    high = highcut / nyq
    b, a = butter(order, [low, high], btype="band")
    return b, a


def butter_bandpass_filter(data, lowcut, highcut, fs, order=4):
    """
    Filters the signal with specified butter-bandpass filter
    :param data: single-channel signal (one-dimensional)
    :param lowcut: . of the filter
    :param highcut: . of the filter
    :param fs: input sampling frequency
    :param order: . of the filter
    :return: filtered signal of same shape
    """
    data = scipy.signal.detrend(data)
    b, a = butter_bandpass(lowcut, highcut, fs, order=order)
    y = lfilter(b, a, data)
    return y


def process_window_spec_ppg(sig, freq, resolution, min_hz, max_hz):
    """
    Processes a window of PPG signal into spectral representation. Custom implementation of STFT.
    The channels are averaged, the signal is downsampled and the FFT is computed. Only amplitudes within relevant
    range are returned.
    :param sig: PPG signal of shape (n_samples, n_channels)
    :param freq: signal frequency (Hz)
    :param resolution: Number of points for the FFT algorithm
    :param min_hz: Lower cutoff frequency for spectrum
    :param max_hz: Higher cutoff frequency for spectrum
    :return: Spectrogram of shape (n_steps, n_freq_bins)
    """
    filt = lambda x: butter_bandpass_filter(x, 0.4, 4, fs=freq, order=4)
    sig = np.apply_along_axis(filt, 0, sig)
    sig = (sig - sig.mean(axis=0)[None, :]) / (sig.std(axis=0)[None, :] + 1e-10)
    sig = sig.mean(axis=-1)  # average over channels if multiple present

    sig = scipy.signal.resample(
        sig, int(len(sig) * 25 / freq)
    )  # resample to 25 hz (optional but faster)
    # do FFT
    if resolution > len(sig):
        sig = np.pad(sig, (0, resolution - len(sig)))
    y = np.abs(rfft(sig, axis=0))
    freq = rfftfreq(len(sig), 1 / 25)
    # extract relevant frequencies
    y = y[(freq > min_hz) & (freq < max_hz)]
    return y


def process_window_spec_acc(sig, freq, resolution: int, min_hz: float, max_hz: float):
    """
    Processes a window of accelerometer signal into spectral representation. Custom implementation of STFT.
    The signal is downsampled, FFT is computed and the channels are averaged. Only amplitudes within relevant
    range are returned.
    :param sig: ACC signal of shape (n_samples, n_channels)
    :param freq: signal frequency (Hz)
    :param resolution: Number of points for the FFT algorithm
    :param min_hz: Lower cutoff frequency for spectrum
    :param max_hz: Higher cutoff frequency for spectrum
    :return: Spectrogram of shape (n_steps, n_freq_bins)
    """
    filt = lambda x: butter_bandpass_filter(x, 0.4, 4, fs=freq, order=4)
    sig = np.apply_along_axis(filt, 0, sig)
    sig = (sig - sig.mean(axis=0)[None, :]) / (sig.std(axis=0)[None, :] + 1e-10)

    sig = scipy.signal.resample(
        sig, int(len(sig) * 25 / freq)
    )  # resample to 25 hz (optional but faster)
    # do FFT
    if resolution > len(sig):
        sig = np.pad(sig, ((0, resolution - len(sig)), (0, 0)))
    y = np.abs(rfft(sig, axis=0))
    freq = rfftfreq(len(sig), 1 / 25)
    # extract relevant frequencies
    y = y[(freq > min_hz) & (freq < max_hz), :]
    return y.mean(axis=-1)


def process_window_time(ppg, ppg_freq, target_freq, filter_lowcut, filter_highcut):
    """
    Prepares time-domain PPG signal as input. Preprocessing consists of filtering, averaging & resampling.
    :param ppg: PPG signal of shape (n_samples, n_channels)
    :param ppg_freq: frequency of signal
    :param target_freq: desired frequency for input
    :param filter_lowcut: lower cutoff frequency for bandpass filtering
    :param filter_highcut: upper cutoff frequency for bandpass filtering
    :return: time-domain signal of shape (n_samples_new,)
    """
    filt = lambda x: butter_bandpass_filter(x, filter_lowcut, filter_highcut, ppg_freq)
    ppg = np.apply_along_axis(filt, 0, ppg)
    ppg = ppg.mean(axis=-1)  # average over channels if multiple present
    ppg = scipy.signal.resample_poly(ppg, target_freq, ppg_freq)
    ppg = np.expand_dims(ppg, -1)
    return ppg



from tensorflow.keras.layers import (
    LSTM,
    Activation,
    BatchNormalization,
    Conv1D,
    Dense,
    Dropout,
    MaxPooling1D,
)


def get_timedomain_backbone(inp, output_shape):
    """
    generates the time backbone, whose architecture corresponds to CorNetLight
    :param model_in: time-domain input
    :param output_shape: what last dimension the bottleneck has
    :return: two branches, one for features and one for weight
    """
    x = inp
    # Block 1
    x = Conv1D(16, kernel_size=10, strides=1, dilation_rate=2, padding="causal")(x)
    x = BatchNormalization()(x)
    x = Dropout(0.1)(x)
    x = Activation("leaky_relu")(x)
    x = MaxPooling1D(4, strides=4)(x)
    # Block 2
    x = Conv1D(16, 10, 1, dilation_rate=2, padding="causal")(x)
    x = BatchNormalization()(x)
    x = Dropout(0.1)(x)
    x = Activation("leaky_relu")(x)
    x = MaxPooling1D(4, strides=4)(x)

    x = LSTM(64, activation="tanh", dropout=0.1, return_sequences=True)(
        x, training=True
    )
    x = LSTM(64, activation="tanh", dropout=0.1, return_sequences=False)(
        x, training=True
    )

    feat_branch = Dense(output_shape, activation="leaky_relu")(x)
    value_branch = Dense(output_shape, activation="leaky_relu")(x)
    return feat_branch, value_branch



""""""
import glob
import logging
import os

import numpy as np
import pandas as pd
from scipy.io import loadmat


def load_dalia(data_dir):
    """
    Functionality to load PPG-DaLiA dataset. Assumes it is stored under data_dir/DaLiA/*
    Downloadable under https://archive.ics.uci.edu/ml/machine-learning-databases/00495/data.zip
    :param data_dir: base directory to data folder
    :return: tuple of (signals, names, ppg_freq, acc_freq
        signals: list of signal tuples (ppg, acc, hr) each representing one session/subject.
                 Signals are assumed to be in channels-last format.
                 Ground truth hr is assumed to be the instantaneous HR in an 8-second window with 2s stride.
        names:   list of strings describing the sessions
        ppg_freq, acc_freq: respective sampling frequencies
    """
    signals, names = [], []
    ppg_freq = 64
    acc_freq = 32
    nsamples = 0
    tpl = os.path.join(*[data_dir, "DaLia", "S*", "S*.pkl"])
    for fname in sorted(glob.glob(tpl)):
        # load
        ds = np.load(fname, allow_pickle=True, encoding="bytes")
        # parse
        acc = ds[b"signal"][b"wrist"][b"ACC"]
        ppg = ds[b"signal"][b"wrist"][b"BVP"]
        hr = ds[b"label"]
        signals.append((ppg, acc, hr))
        names.append("DaLia_" + os.path.split(fname)[1][:-4])
        nsamples += len(ppg) / (60 * ppg_freq)

    assert nsamples > 0, r"Did not find any files matching path %s" % tpl

    logging.info(
        r"Loaded DaLiA (signal length: %d min, number of sessions: %d)"
        % (nsamples, len(signals))
    )
    return signals, names, ppg_freq, acc_freq


def load_wesad(data_dir):
    """
    Functionality to load WESAD dataset. Assumes it is stored under data_dir/WESAD/*
    Downloadable under https://uni-siegen.sciebo.de/s/HGdUkoNlW1Ub0Gx/download
    :param data_dir: base directory to data folder
    :return: tuple of (signals, names, ppg_freq, acc_freq
        signals: list of signal tuples (ppg, acc, hr) each representing one session/subject.
                 Signals are assumed to be in channels-last format.
                 Ground truth hr is assumed to be the instantaneous HR in an 8-second window with 2s stride.
        names:   list of strings describing the sessions
        ppg_freq, acc_freq: respective sampling frequencies
    """
    signals, names = [], []
    ppg_freq = 64
    acc_freq = 32
    nsamples = 0
    tpl = os.path.join(*[data_dir, "WESAD", "S*", "S*.pkl"])
    for fname in sorted(glob.glob(tpl)):
        # load
        ds = np.load(fname, allow_pickle=True, encoding="bytes")
        name = os.path.split(fname)[1][:-4]
        # parse
        ppg = ds[b"signal"][b"wrist"][b"BVP"]
        acc = ds[b"signal"][b"wrist"][b"ACC"]
        hr = np.loadtxt(
            os.path.join(*[data_dir, "WESAD", "GeneratedLabels", f"HR_{name}.csv"]),
            delimiter=",",
        )
        signals.append((ppg, acc, hr))
        names.append("WESAD_" + name)
        nsamples += len(ppg) / (60 * ppg_freq)

    assert nsamples > 0, r"Did not find any files matching path %s" % tpl

    logging.info(
        r"Loaded WESAD (signal length: %d min, number of sessions: %d)"
        % (nsamples, len(signals))
    )
    return signals, names, ppg_freq, acc_freq


def load_bami_1(data_dir):
    """
    Functionality to load the BAMI-1 dataset. Assumes it is stored under data_dir/BAMI/BAMI-1*
    Downloadable jointly with BAMI-2 under
    https://github.com/HeewonChung92/CNN_LSTM_HeartRateEstimation/archive/refs/heads/master.zip
    :param data_dir: base directory to data folder
    :return: tuple of (signals, names, ppg_freq, acc_freq
        signals: list of signal tuples (ppg, acc, hr) each representing one session/subject.
                 Signals are assumed to be in channels-last format.
                 Ground truth hr is assumed to be the instantaneous HR in an 8-second window with 2s stride.
        names:   list of strings describing the sessions
        ppg_freq, acc_freq: respective sampling frequencies
    """
    signals, names = [], []
    ppg_freq = 50
    acc_freq = 50
    nsamples = 0
    tpl = os.path.join(*[data_dir, "BAMI", "BAMI-1", "BAMI1_*.mat"])
    for fname in sorted(glob.glob(tpl)):
        # load
        mat = loadmat(fname)
        # parse
        ppg = np.array(mat["rawPPG"]).T
        acc = np.array(mat["rawAcc"]).T
        hr = np.array(mat["bpm_ecg"]).flatten()
        length = int((len(ppg) - 6 * 50) / (2 * 50))
        hr = hr[-length:]  # there is a sequence with one label too much... wonder why?

        signals.append((ppg, acc, hr))
        names.append(os.path.split(fname)[1][:-4])
        nsamples += len(ppg) / (60 * ppg_freq)

    assert nsamples > 0, r"Did not find any files matching path %s" % tpl

    logging.info(
        r"Loaded BAMI-1 (signal length: %d min, number of sessions: %d)"
        % (nsamples, len(signals))
    )
    return signals, names, ppg_freq, acc_freq


def load_bami_2(data_dir):
    """
    Functionality to load the BAMI-2 dataset. Assumes it is stored under data_dir/BAMI/BAMI-2*
    Downloadable jointly with BAMI-1 under
    https://github.com/HeewonChung92/CNN_LSTM_HeartRateEstimation/archive/refs/heads/master.zip
    :param data_dir: base directory to data folder
    :return: tuple of (signals, names, ppg_freq, acc_freq
        signals: list of signal tuples (ppg, acc, hr) each representing one session/subject.
                 Signals are assumed to be in channels-last format.
                 Ground truth hr is assumed to be the instantaneous HR in an 8-second window with 2s stride.
        names:   list of strings describing the sessions
        ppg_freq, acc_freq: respective sampling frequencies
    """
    signals, names = [], []
    ppg_freq = 50
    acc_freq = 50
    nsamples = 0
    tpl = os.path.join(*[data_dir, "BAMI", "BAMI-2", "BAMI2_*.mat"])
    for fname in sorted(glob.glob(tpl)):
        # load
        mat = loadmat(fname)
        # parse
        ppg = np.array(mat["rawPPG"]).T
        acc = np.array(mat["rawAcc"]).T
        hr = np.array(mat["bpm_ecg"]).flatten()
        length = int((len(ppg) - 6 * 50) / (2 * 50))
        length2 = len(hr) * (2 * 50) + 6 * 50
        hr = hr[-length:]  # there is a sequence with one label too much... wonder why?
        ppg = ppg[-length2:]  # one sequence is one label short... wonder why?
        acc = acc[-length2:]
        signals.append((ppg, acc, hr))
        names.append(os.path.split(fname)[1][:-4])
        nsamples += len(ppg) / (60 * ppg_freq)

    assert nsamples > 0, r"Did not find any files matching path %s" % tpl

    logging.info(
        r"Loaded BAMI-2 (signal length: %d min, number of sessions: %d)"
        % (nsamples, len(signals))
    )
    return signals, names, ppg_freq, acc_freq


def load_ieee(data_dir, load_train=True, load_extra=False, load_test=False):
    """
    Functionality to load IEEE train and test datasets. Assumes it is stored under data_dir/IEEE/*
    Original link became unreachable. Downloadable in a NEW FORMAT under https://zenodo.org/record/3902710#.ZGKi-3ZBy3A
    :param data_dir: base directory to data folder
    :param load_train: whether to load the training sequences
    :param load_extra: whether to load the 13th training sequence
    :param load_test: whether to load the test sequences
    :return: tuple of (signals, names, ppg_freq, acc_freq
        signals: list of signal tuples (ppg, acc, hr) each representing one session/subject.
                 Signals are assumed to be in channels-last format.
                 Ground truth hr is assumed to be the instantaneous HR in an 8-second window with 2s stride of matching length.
        names:   list of strings describing the sessions
        ppg_freq, acc_freq: respective sampling frequencies
    """
    ppg_freq, acc_freq = 125, 125
    signals, names = [], []
    nsamples = 0
    # prepare filenames
    train_tpl = os.path.join(
        *[data_dir, "IEEE", "Training_data", "DATA_*_TYPE[0-9][0-9].mat"]
    )
    extra_tpl = os.path.join(
        *[data_dir, "IEEE", "Extra_TrainingData", "DATA_*_T[0-9]*.mat"]
    )
    test_tpl = os.path.join(*[data_dir, "IEEE", "IEEE", "TestData", "TEST_S*_T*.mat"])

    train_names = (glob.glob(train_tpl) if load_train else []) + (
        glob.glob(extra_tpl) if load_extra else []
    )
    test_names = glob.glob(test_tpl) if load_test else []
    fnames = sorted(train_names) + sorted(test_names)

    for fname in fnames:
        mat = loadmat(fname)
        # get the labels
        if "Extra_TrainingData" in fname:
            bpm_file = fname.replace("DATA", "BPM")
        elif "Training" in fname:
            bpm_file = fname.replace(".mat", "_BPMtrace.mat")
        else:
            bpm_file = fname.replace("TEST", "True")

        hr = loadmat(bpm_file)["BPM0"].flatten()
        df = pd.DataFrame(mat["sig"].swapaxes(0, 1))

        df.columns = ["ecg", "ppg1", "ppg2", "acc1", "acc2", "acc3"]

        ppg = np.stack([df["ppg1"], df["ppg2"]], axis=-1)
        acc = np.stack([df["acc1"], df["acc2"], df["acc3"]], axis=-1)

        signals.append((ppg, acc, hr))
        names.append("IEEE_"+os.path.split(fname)[1][:-4])
        nsamples += len(ppg) / (60 * ppg_freq)

    assert nsamples > 0, r"Did not find any files matching paths %s %s %s" % (
        train_tpl,
        extra_tpl,
        test_tpl,
    )

    logging.info(
        r"Loaded IEEE (signal length: %d min, number of sessions: %d)"
        % (nsamples, len(signals))
    )
    logging.info((r"The following files were read: %s" % ("\n".join(["\n"] + fnames))))
    return signals, names, ppg_freq, acc_freq



class InputConfig:
    """
    Configuration for input to network.

    This class defines the parameters for windowing the input data sequences. It specifies the
    window size and the stride between consecutive windows.
    """
    WINSIZE = 8  # Length of the input window in seconds.
    STRIDE = 2   # Stride between consecutive windows in seconds.